using System;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Text;
using System.Threading;
using System.Security.Cryptography;
using System.Collections;

namespace P2PBackupClient
{
	/// <summary>
	/// Summary description for User.
	/// </summary>
	public class User
	{
		private string userName;
		private string password;
		private String pubKey;
		private Socket connection;
		private StreamWriter writer;
		private StreamReader reader;
		public ArrayList sessionsConnect = new  ArrayList();//used when user connects to other clients
		public ArrayList sessionsListen = new  ArrayList();//used when other clients connects to user
		private bool run = false;
		private long shareSize;
		private string sharePath;
		private RSACryptoServiceProvider keyPairCrypto = null;
		public FileHandle fileHand = new FileHandle();
		private BFileList bfl = null;
		private string accessible;
		
		public delegate void UpdateGUIHandler (string command, string param);
		public event UpdateGUIHandler UpdateGUIEvent;//Client.UpdateGUI()

		public User()
		{
			connection = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
			GetKeyPair(out keyPairCrypto);
			string [] sizePath = fileHand.GetSharePathSize();
			if(sizePath[0] != null)
			{
				shareSize = Int64.Parse(sizePath[0]);
				sharePath = sizePath[1];
			}
		}

		public string Username
		{
			set { userName = value; }
			get { return userName; }
		}

		public string Password
		{
			set { password = value; }
			get { return password; }
		}

		public bool Connected
		{
			get{ return connection.Connected; }
		}

		public string SharePath
		{
			get { return sharePath; }
		}

		public long ShareSize
		{
			get { return shareSize; }
		}

		public  RSACryptoServiceProvider KeyPairCrypto
		{
			get { return keyPairCrypto; }
		}

		public bool Run
		{
			get {return run;}
		}

		/// <summary>
		/// Connects the user to the hub and creates Reader and Writer for communication.
		/// </summary>
		/// <param name="ip">Hub IP</param>
		/// <param name="port">Hub Port</param>
		public void ConnectToHub(string ip, int port)
		{
			try
			{
				IPAddress addr = IPAddress.Parse(ip);
				IPEndPoint dest = new IPEndPoint(addr, port);
				connection.Connect(dest);
				NetworkStream netstream = new NetworkStream(connection);
				reader = new StreamReader(netstream,Encoding.Unicode);
				writer = new StreamWriter(netstream,Encoding.Unicode);
			}
			catch(SocketException)
			{
				MessageBox.Show("Unable to connect to hub!");
			}
			catch(ObjectDisposedException)
			{
				// if the user, for some reason, has been logged out
				// and the user is logging in again
				// create a new socket and call ConnectToHub again
				connection = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
				ConnectToHub(ip, port);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.ConnectToHub"); 
			}
		}
		
		/// <summary>
		/// Sets the client in listening mode.
		/// </summary>
		public void StartListening()
		{
			try
			{
				Thread t = new Thread(new ThreadStart(Listen));
				t.IsBackground = true;
				t.Start(); 
				run = true;
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.StartListening"); 
			}
		}

		/// <summary>
		/// Breaks the listening mode and disconnects the client from the hub.
		/// </summary>
		public void Disconnect(bool closing)
		{
			try
			{
				run = false;
				if (writer != null)
					writer.Close ();

				Thread.Sleep(1024);
				if(this.connection != null && this.connection.Connected)
				{
					this.connection.Shutdown(SocketShutdown.Both);
					this.connection.Close();
				}
				if(UpdateGUIEvent != null)
					UpdateGUIEvent("OUT", "");
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.Disconnect"); 
			}
		}

		/// <summary>
		/// Tells session to connect to the other client
		/// </summary>
		/// <param name="client2">the session that handles all communication with the other client</param>
		public void ConnectToClient(Session client2)
		{
			client2.ConnectToClient();
		}

		/// <summary>
		/// Sends the username to the hub for login
		/// </summary>
		public void Login()
		{
			try
			{
				writer.WriteLine("LOG " + userName); 	
				writer.Flush();
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.Login"); 
			}
		}
		
		/// <summary>
		///  Hashes the password entered by user and store in user.password
		/// </summary>
		/// <param name="password">user's password</param>
		public void HashPassword(string password)
		{
			string hashed;
			hashed = HashString(password);	
			this.password = hashed;
		}

		/// <summary>
		/// Does the preparations for sending a backup.
		/// </summary>
		/// <param name="nameBackup">the name for backup</param>
		/// <param name="pathBackup">directory to backup</param>
		/// <param name="pathTemp">directory for temporary files</param>
		/// <param name="saveLocal">tells if local copy should be saved, true to save</param></param>
		public void PrepareBackup(string nameBackup, string pathBackup, string pathTemp, bool saveLocal)
		{
			DirectoryInfo dInf = new DirectoryInfo(pathBackup);
			bfl = new BFileList();
			bfl.SaveBackupLocal = saveLocal;
			bfl.GetFileInfo(dInf,pathTemp);//get all fileinfos
			DateTime time = DateTime.Now;
			string nb = this.userName + "_" + time.ToShortDateString() + "_" + time.ToShortTimeString().Remove(2,1) + "_" + nameBackup;
			fileHand.CreateBackupIndex(nb);//creates indexfile for backup
			bfl.IndexFileName = nb + ".p2p";
			bfl.GetFileToEncrypt(nb, keyPairCrypto);
			string size = bfl.GetBackupSize();
			SendBackupToHub(size);//sends BCK, backup size
		}

		/// <summary>
		/// Tells the hub that client/user wants to take a backup
		/// </summary>
		/// <param name="size">total size of backup</param>
		private void SendBackupToHub(string size)
		{
			try
			{
				writer.WriteLine("BCK " + size);
				writer.Flush();
				if (UpdateGUIEvent != null)
				{
					long total = Int64.Parse(size);
					UpdateGUIEvent("initprbBackup", total.ToString());
				}
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.SendBackupToHub"); 
			}
		}

		/// <summary>
		/// Request a destinations from the hub for BFile that has not been sent.
		/// Also checks if all files has been sent.
		/// </summary>
		private void SendPut()
		{
			bool allBackupFilesSent = true;
			bool sendDoneBackup = false;
			foreach (BFile bf in bfl)
			{
				if(bf.IndexFile &&  bf.Sent)
					sendDoneBackup = true;
				if(!bf.Sent)
				{
					try
					{
						writer.WriteLine("PUT " + bf.SubstituteFileName + " " + bf.FileSize);
						writer.Flush();
					}
					catch(IOException)
					{
						this.Disconnect(false);
					}
					catch(Exception ex)
					{
						MessageBox.Show(ex.Message,"User.SendPut"); 
					}
					allBackupFilesSent = false;
					break;
				}
			}
			if (allBackupFilesSent && !sendDoneBackup)
			{
				SendIndexFile();	
			}
			if (sendDoneBackup)
			{
				SendDoneBackup();
				bfl = null;
			}
		}

		/// <summary>
		/// Tells the hub that a file has been back uped to a client.
		/// </summary>
		/// <param name="bf">the file that has been back uped</param>
		private void SendDoneFile(BFile bf)
		{
			try
			{
				writer.WriteLine("DFI " + bf.SubstituteFileName + " " + bf.ClientName);
				writer.Flush();
				if(UpdateGUIEvent != null)
					UpdateGUIEvent("prbBackup", bf.FileSize.ToString());
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.SendDoneFile"); 
			}
		}

		/// <summary>
		/// Tells the hub that backup is done.
		/// </summary>
		private void SendDoneBackup()
		{	
			if(bfl.SaveBackupLocal == false)
				bfl.DeleteEncryptedFiles();
			UpdateGUIEvent("DoneBackup","");
			try
			{
				writer.WriteLine("DBU");
				writer.Flush();
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.SendDoneBackup"); 
			}
		}

		/// <summary>
		/// Tells the hub that user/client is ready for another client to connect.
		/// </summary>
		/// <param name="clientName">the other clients username</param>
		private void SendAccept(string clientName)
		{
			try
			{
				writer.WriteLine("ACC " + clientName);
				writer.Flush();
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.SendAccept"); 
			}
		}

		/// <summary>
		/// Does the preparations for a recovery.
		/// </summary>
		/// <param name="backupName">the name of the backup, in this case the same as the name for indexfile</param>
		public void PrepareRecovery(string backupName)
		{
			Directory.CreateDirectory("Temp");
			bfl = fileHand.GetFromIndexFile(backupName);
			long total = Int64.Parse(bfl.GetBackupSize());
			UpdateGUIEvent("initprbRecovery", total.ToString());
			SendRecoveryToHub();
		}

		/// <summary>
		/// Tells the hub that client/user wants to fetch file from other client.
		/// When all files are fetched, they are decrypted and temporary folder is deleted.
		/// </summary>
		private void SendRecoveryToHub()
		{
			bool allFilesFetched = true;
			foreach (BFile bf in bfl)
			{
				if(bf.Fetched == false)
				{
					try
					{
						writer.WriteLine("REC " + bf.ClientName);
						writer.Flush();
					}
					catch(IOException)
					{
						this.Disconnect(false);
					}
					catch(Exception ex)
					{
						MessageBox.Show(ex.Message,"User.SendRecoveryToHub"); 
					}
					allFilesFetched = false;;
					break;
				}
			}
			if (allFilesFetched == true)
			{
				long total = Int64.Parse(bfl.GetBackupSize());
				bfl.GetFileToDecrypt(keyPairCrypto);
				try
				{
					Directory.Delete("Temp", true);
				}
				catch (Exception ex)
				{
					MessageBox.Show(ex.Message, "User.SendRecoveryToHub");
				}
				bfl = null;
			}
		}

		/// <summary>
		/// Saves the size and path of the share and sends the size to the hub
		/// </summary>
		/// <param name="size">size of the share</param>
		/// <param name="path">path of the share</param>
		public void SendShare(string size, string path)
		{
			shareSize = Int64.Parse(size);
			shareSize = shareSize * 1048576;
			size = shareSize.ToString();
			sharePath = path;

			try
			{
				writer.WriteLine("SHA " + size);
				writer.Flush();
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.SendShare"); 
			}
		}

		/// <summary>
		/// Creates an asymmetric keypair and saves it to file.
		/// Saves public key in User.
		/// </summary>
		public void CreateKeyPair()
		{	
			if(keyPairCrypto == null)
			{
				char[]  keyData;    // Key data as a Char array.
				byte[]  keyByte;    // Converted key data.
				//UpdateGUIEvent("progBar","");
				// Generate a key pair.
				keyPairCrypto = new RSACryptoServiceProvider(2048);
				if (UpdateGUIEvent != null)
					UpdateGUIEvent("prbKeyPair","20");
				// Convert the key data for storage, chararray to bytearray
				keyData = keyPairCrypto.ToXmlString(true).ToCharArray();
				if (UpdateGUIEvent != null)
					UpdateGUIEvent("prbKeyPair","40");
				keyByte = new Byte[keyData.Length];
				for (int Counter = 0; Counter < keyData.Length; Counter++)
					keyByte[Counter] = Convert.ToByte(keyData[Counter]);
				if (UpdateGUIEvent != null)
					UpdateGUIEvent("prbKeyPair","60");
				fileHand.SaveKeyPair(keyByte);//Saves key pair to file
				if (UpdateGUIEvent != null)
					UpdateGUIEvent("prbKeyPair","80");
				pubKey = keyPairCrypto.ToXmlString(false);
				if (UpdateGUIEvent != null)
					UpdateGUIEvent("prbKeyPair","100");
			}	
		}

		/// <summary>
		/// Creates a BFile for the index file, encrypt it and add it to BFileList.
		/// Then tells the hub that a indexfile is to be back uped.
		/// </summary>
		private void SendIndexFile()
		{
			string path = Application.ExecutablePath;
			path = path.Substring(0, path.LastIndexOf(@"\") + 1);
			DirectoryInfo dInf = new DirectoryInfo(path);
			string search = bfl.IndexFileName;
			FileInfo[] fi = dInf.GetFiles(search);
			BFile indexFile = new BFile(bfl.IndexFileName);
			indexFile.IndexFile = true;
			indexFile.SubstituteFileName = bfl.IndexFileName;
			indexFile.OriginalFileName = fi[0].FullName;
			indexFile.FileSize = fi[0].Length;
			foreach(BFile bf in bfl)
			{	
				indexFile.PathTemp = bf.PathTemp;
				break;
			}
			indexFile.EncryptFile(keyPairCrypto);
			bfl.Add(indexFile);

			try
			{
				writer.WriteLine("PIX " + indexFile.SubstituteFileName + " " + indexFile.FileSize.ToString());
				writer.Flush();
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.SendIndexFile"); 
			}

		}

		/// <summary>
		/// Sends the public key to the hub
		/// </summary>
		public void SendPublicKey()
		{
			try
			{
				if (pubKey != null)
				{
					writer.WriteLine("KEY " + pubKey);
					writer.Flush();
				}
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.SendPublicKey"); 
			}
		}

		/// <summary>
		/// Gets the key pair from file.
		/// Saves public key in User.
		/// </summary>
		/// <param name="keyPair">the key pair for user</param>
		private void GetKeyPair(out RSACryptoServiceProvider keyPair)
		{
			byte[] keyByte;
			keyByte = fileHand.GetKeyPair();

			if(keyByte != null)
			{
				keyPair = new RSACryptoServiceProvider();

				StringBuilder sb = new StringBuilder(keyByte.Length);
				for (int Counter = 0; Counter < keyByte.Length; Counter++)
					sb.Append(Convert.ToChar(keyByte[Counter]));
				keyPair.FromXmlString(sb.ToString());
		
				pubKey = keyPair.ToXmlString(false);

			}
			else
				keyPair = null;		
		}

		/// <summary>
		/// Listens to hub and send all incoming messages to user.Decode.
		/// </summary>
		private void Listen()
		{
			try
			{				
				using (reader)
				{
					while(run)
					{
						Decode(reader.ReadLine()); 
					}
				}
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.Listen"); 
			}
		}

		/// <summary>
		/// Decodes all messages that arrive from the hub.
		/// </summary>
		/// <param name="message">arrived message</param>
		private void Decode(string message)
		{
			try
			{
				if (message == null)
				{
					Disconnect(false);
					return;
				}
				string type="";
				char[] separator = {' '};
				string[] decoded = message.Split(separator);
				if(decoded[0].Length==3)
					type = decoded[0];
				
				switch(type)
				{
						//ALPHABETICAL!!!
					case "ACS":
						if(decoded.Length == 2 && UpdateGUIEvent != null)
						{
							accessible = decoded[1];
							UpdateGUIEvent("ACS",decoded[1]);
						}
						break;
					case "CHA":
						if(decoded.Length == 3)
							Respond(decoded[1],decoded[2]);
						break;
					case "DST"://Destination for file
						if(decoded.Length == 5)
							SendBackup(decoded[1],decoded[2],decoded[3],decoded[4]);
						break;
					case "LET"://Destination for file
						if(decoded.Length == 4)
							SendRecovery(decoded[1],decoded[2],decoded[3]);
						break;
					case "RCV"://Receive file
						if(decoded.Length == 4)
							ReceiveBackup(decoded[1], decoded[2], decoded[3]);
						break;
					case "SRC"://Receive file
						if(decoded.Length == 4)
							GetRecovery(decoded[1], decoded[2], decoded[3]);
						break;
						//NUMMERORDNING !!!
					case "201": //Verification klient-hub ok
						if(decoded.Length == 1 && UpdateGUIEvent != null)
						{
							UpdateGUIEvent("201","");
						}
						break;
					case "202": //OK to take backup
						if(decoded.Length == 1)
						{
							SendPut();	
						}
						break;
					case "208": //Share ok
						if(decoded.Length == 1 && UpdateGUIEvent != null)
						{
							UpdateGUIEvent("208","");
						}
						break;
					case "209": //Share ok
						if(decoded.Length == 1 && UpdateGUIEvent != null)
						{
							UpdateGUIEvent("209","");
						}
						break;
					case "215":
						if(decoded.Length == 1)
							SendPut();
						break;
					case "301"://Verification klient-hub ok, but need share size
						if(decoded.Length == 1 && UpdateGUIEvent != null)
						{
							UpdateGUIEvent("301","");
						}
						break;
					case "302"://Verification klient-hub ok or Share ok, but need public key
						if(decoded.Length == 1 && UpdateGUIEvent != null)
						{
							UpdateGUIEvent("302","");
						}
						break;
					case "401"://User doesn't exist
						if(decoded.Length == 1 && UpdateGUIEvent != null)
						{
							UpdateGUIEvent("401","");
						}
						break;
					case "402"://Verification client-hub did not succeed
						if(decoded.Length == 1 && UpdateGUIEvent != null)
						{
							UpdateGUIEvent("402","");
						}
						break;
					case "403"://Not enough space for backup
						if(decoded.Length == 1 && UpdateGUIEvent != null)
						{
							bfl.DeleteEncryptedFiles();
							fileHand.DeleteIndexFile(bfl.BackupName);
							bfl = null;
							UpdateGUIEvent("403","");
						}
						break;
					case "404"://Share is smaller than used space
						if(decoded.Length == 1)
						{
							UpdateGUIEvent("404","");
						}
						break;
					case "501"://IP-address blocked
						if(decoded.Length == 1)
						{
							
						}
						break;
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.Decode"); 
			}
		}

		/// <summary>
		/// Hashes a string
		/// </summary>
		/// <param name="data">string to hash</param>
		/// <returns>hashed string</returns>
		private string HashString(string data)
		{
			byte[] toHash = Encoding.ASCII.GetBytes(data);
			byte [] hashedResult;

			SHA1Managed hashAlg = new SHA1Managed();
			hashedResult = hashAlg.ComputeHash(toHash);
			return ByteArrayToString(hashedResult);
		}

		/// <summary>
		/// Takes a bytearray and creates a string.
		/// </summary>
		/// <param name="toConvert">byte[] to convert</param>
		/// <returns>string representation of byte[]</returns>
		private string ByteArrayToString(byte [] toConvert)
		{
			StringBuilder sb = new StringBuilder(toConvert.Length);
			for (int i = 0; i < toConvert.Length - 1; i++)
			{
				sb.Append(toConvert[i].ToString("X"));
			}
			return sb.ToString();
		}
		
		/// <summary>
		/// Takes time and random number received from the hub. 
		/// Hashes it together with password.
		/// Sends result to the hub.
		/// </summary>
		/// <param name="time">time stamp</param>
		/// <param name="no">random number</param>
		private void Respond(string time, string no)
		{
			try
			{
				string hashed;
				string data = this.password + time + no;
				hashed = HashString(data);			
				writer.WriteLine("RSP " + hashed);
				writer.Flush();
			}
			catch(IOException)
			{
				this.Disconnect(false);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.Respond"); 
			}
		}

		/// <summary>
		/// Starts listening for and accepts a connection from other client
		/// </summary>
		private void StartListeningForClient(string clientName)
		{
			try
			{
				Socket socket = new Socket(AddressFamily.InterNetwork,SocketType.Stream,ProtocolType.Tcp);
				IPEndPoint src = new IPEndPoint(IPAddress.Any, 52562);
				socket.Bind(src);
				SendAccept(clientName);
				socket.Listen(1);
				Socket clientSession = socket.Accept();
				string [] ip = new String[2];
				ip = clientSession.RemoteEndPoint.ToString().Split(':');

				// compares the sockets ip with the ones in the arraylist
				foreach (Session session in sessionsListen)
				{
					// if ip:s are the same, start listening
					if (ip[0] == session.ClientIp)
					{
						session.Client = clientSession;
						session.StartListening();
						socket.Close();
						Thread t = new Thread(new ThreadStart(this.RemoveConnections));
						t.IsBackground = true;
						t.Start();
						break;
					}
				}		
			}
			catch(ObjectDisposedException)
			{
				// socket closed when user logged out, do nothing
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"User.StartListeningForClients");
			}
			
		}

		/// <summary>
		/// Removes session from session list
		/// </summary>
		/// <param name="s">session to remove</param>
		/// <param name="sessionType">sessionType to remove</param>
		private void RemoveSession(Session s,int sessionType)
		{
			if(sessionType == 1 || sessionType == 2)
				sessionsConnect.Remove(s);
			else
				sessionsListen.Remove(s);
		}

		/// <summary>
		/// If file has been sent, SendPut is called.
		/// </summary>
		/// <param name="sent">true if file has been sent</param>
		/// <param name="bFSent">the file that has been sent</param>
		private void FileSent(bool sent,BFile bFSent)
		{
			if (sent == true)
			{
				if(!bFSent.IndexFile)
					fileHand.SaveToIndexFile(bFSent, bfl.BackupName);
				foreach (BFile bf in bfl)
				{
					if(bf.OriginalFileName == bFSent.OriginalFileName)
					{
						bf.Sent = true;
						try
						{
							FileAttributes fa = File.GetAttributes(bf.OriginalFileName);
							if((fa & FileAttributes.Archive) == FileAttributes.Archive)
							{//Removes the archive attribute from file
								fa = fa ^ FileAttributes.Archive;
								File.SetAttributes(bf.OriginalFileName,fa );
							}
						}
						catch(Exception ex)
						{
							MessageBox.Show(ex.Message,"User.FileSent");
						}

						SendDoneFile(bf);
					}	
				}
			}
		}

		/// <summary>
		/// If file is recovery-file is received SendRecoveryToHub is called
		/// </summary>
		/// <param name="received">true if file has been received</param>
		/// <param name="bFReceived">the file that has been received</param>
		private void FileReceived(bool received, BFile bFReceived)
		{
			if (received == true)
			{
				foreach (BFile bf in bfl)
				{
					if(bf.OriginalFileName == bFReceived.OriginalFileName)
					{
						bf.Fetched = true;
						UpdateGUIEvent("prbRecovery", bf.FileSize.ToString());
						SendRecoveryToHub();
					}	
				}
			}
		}

		/// <summary>
		/// Creates a session for sending a backup
		/// </summary>
		/// <param name="fN">fileName(file to be sent)</param>
		/// <param name="cIp">client2 Ip</param>
		/// <param name="cN">client2 username</param>
		/// <param name="cK">client2 public key</param>
		private void SendBackup(string fN, string cIp, string cN, string cK)
		{
			foreach(BFile bf in bfl)
			{
				if(bf.SubstituteFileName ==  fN)
				{
					Session client2 = new Session(1,this, cIp,cN,cK,this.keyPairCrypto,bf);
					client2.RemoveSessionEvent += new P2PBackupClient.Session.RemoveSessionHandler(this.RemoveSession);
					client2.FileSentEvent += new P2PBackupClient.Session.FileSentHandler(this.FileSent);
					client2.FileReceivedEvent += new P2PBackupClient.Session.FileReceivedHandler(this.FileReceived);
					sessionsConnect.Add(client2);
					ConnectToClient(client2);
					break;
				}
			}	
		}

		/// <summary>
		/// Creates a session for receiving a backup
		/// </summary>
		/// <param name="cIp">client1 IP</param>
		/// <param name="cN">client1 IP</param>
		/// <param name="cK">client1 public key</param>
		private void ReceiveBackup(string cIp, string cN, string cK)
		{
			Session client1 = new Session(3,this, cIp, cN, cK, this.keyPairCrypto,this.SharePath);
			client1.RemoveSessionEvent += new P2PBackupClient.Session.RemoveSessionHandler(this.RemoveSession);
			client1.FileSentEvent += new P2PBackupClient.Session.FileSentHandler(this.FileSent);
			client1.FileReceivedEvent += new P2PBackupClient.Session.FileReceivedHandler(this.FileReceived);
			sessionsListen.Add(client1);
			StartListeningForClient(cN);
		}

		/// <summary>
		/// Creates a session for receiving a recovery
		/// </summary>
		/// <param name="cIp">client2 IP</param>
		/// <param name="cN">client2 IP</param>
		/// <param name="cK">client2 public key</param>
		private void GetRecovery(string cIp, string cN, string cK)
		{
			foreach(BFile bf in bfl)
			{
				if(bf.ClientName ==  cN && bf.Fetched == false)
				{
					Session client2 = new Session(2,this, cIp,cN,cK,this.keyPairCrypto,bf);
					client2.RemoveSessionEvent += new P2PBackupClient.Session.RemoveSessionHandler(this.RemoveSession);
					client2.FileSentEvent += new P2PBackupClient.Session.FileSentHandler(this.FileSent);
					client2.FileReceivedEvent += new P2PBackupClient.Session.FileReceivedHandler(this.FileReceived);
					sessionsConnect.Add(client2);
					ConnectToClient(client2);
					break;
				}
			}	
		}

		/// <summary>
		/// Creates a session for sending a recovery
		/// </summary>
		/// <param name="cIp">client1 IP</param>
		/// <param name="cN">client1 IP</param>
		/// <param name="cK">client1 public key</param>
		private void SendRecovery(string cIp, string cN, string cK)
		{
			Session client1 = new Session(4,this, cIp, cN, cK, this.keyPairCrypto,this.SharePath);
			client1.RemoveSessionEvent += new P2PBackupClient.Session.RemoveSessionHandler(this.RemoveSession);
			client1.FileSentEvent += new P2PBackupClient.Session.FileSentHandler(this.FileSent);
			client1.FileReceivedEvent += new P2PBackupClient.Session.FileReceivedHandler(this.FileReceived);
			sessionsListen.Add(client1);
			StartListeningForClient(cN);
		}

		/// <summary>
		/// currently not used
		/// </summary>
		private void RemoveConnections()
		{
		}
	}
}
