using System;
using System.Net.Sockets;
using System.Net;
using System.IO;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Security.Cryptography;

namespace P2PBackupHub
{
	/// <summary>
	/// Holds all information about connected user
	/// </summary>
	public class User
	{
		private long accessible;
		private int action = 0;				// 0 = default	1 = backup	2 = recovery
		private int attempt;
		private long available;
		private Socket connection;
		private string filename;
		private long filesize;
		private string ip;
		private bool locked;
		private string password;
		private string pubKey;
		private StreamReader reader;
		private bool run = false;
		private long share;
		private int uid;
		private string userName;
		private string verificationHash;
		private bool verified = false;
		private string version = "";
		private StreamWriter writer;

		private DBHandle dbhandle;	
		
		// Hub.AcceptClient listens
		public delegate void AddUserHandler(User user);
		public event AddUserHandler AddUserEvent;
		// UserList.RemoveUser listens
		public delegate void RemoveUserHandler(User user);
		public event RemoveUserHandler RemoveUserEvent;
		// Hub.AcceptClient listens
		public delegate void LoggHandler (string user, bool received, string message);
		public event LoggHandler LoggEvent;
		// UserList.CalculateDestination listens
		public delegate User CalculateDestinationHandler(User user);
		public event CalculateDestinationHandler CalculateDestinationEvent;
		// UserList.GetClient listens
		public delegate User GetClientHandler(string userName);
		public event GetClientHandler GetClientEvent;
		
		public Socket Connection
		{
			get { return connection; }
		}

		public string UserName
		{
			set { userName = value; }
			get { return userName; }
		}

		public string Password
		{
			set { password = value; }
		}

		public double Available
		{
			get { return available; }
		}

		public double Accessible
		{
			get { return accessible; }
		}

		public double Filesize
		{
			get { return filesize; }
		}

		public User ()
		{

		}

		public User (string name, string pass, string key, long share, long access, long avail, bool locked, int uid)
		{
			this.userName = name;
			this.password = pass;
			this.pubKey = key;
			this.share = share;
			this.accessible = access;
			this.available = avail;
			this.locked = locked;
			this.uid = uid;
		}

		public User (string name)
		{
			dbhandle = new DBHandle();
			this.userName = name;
		}

		public User (string name, string pass)
		{
			dbhandle = new DBHandle();
			this.userName = name;
			this.password = pass;
		}

		public User(Socket connection)
		{
			try
			{
				dbhandle = new DBHandle();
				this.connection = connection;
				NetworkStream stream = new NetworkStream(this.connection);
				writer = new StreamWriter(stream, Encoding.Unicode);
				reader = new StreamReader(stream, Encoding.Unicode);
				if(LoggEvent != null)
					LoggEvent(this.userName, false, "VER "+version);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.User"); 
			}
		}

		/// <summary>
		/// Decodes incoming messages from client
		/// </summary>
		/// <param name="message">incoming message from client</param>
		private void Decode(string message)
		{
			try
			{
				if (message == null) 
					return;
				if (LoggEvent != null)
					LoggEvent(this.userName,true,message);
				string type="";
				char[] separator = {' '};
				string[] decoded = message.Split(separator);
				if(decoded[0].Length == 3)
					type = decoded[0];
				
				switch(type)
				{
						// ALPHABETICAL
					case "ACC":
						if((decoded.Length == 2) && (verified))
							GetClient(decoded[1]);
						break;
					case "BCK":
						if((decoded.Length == 2) && (verified))
							Backup(decoded[1]);
							break;
					case "DBU":
						if((decoded.Length == 1) && (verified))
							SendAvailableSpace();
						break;
					case "DFI":
						if ((decoded.Length == 3) && (verified))
							FileDone(decoded[1], decoded[2]);
						break;
					case "KEY":
						if((decoded.Length == 2) && (verified))
							SaveKey(decoded [1]);
						break;
					case "LOG":
						if(decoded.Length == 2)
							UserExists(decoded[1]);
						break;
					case "NST":
						if((decoded.Length == 2) && (verified))
							GetDestination(filename, filesize);
						break;
					case "OFF":
						if((decoded.Length == 2) && (verified))
							GetDestination(filename, filesize);
						break;
					case "PIX":
						if((decoded.Length == 3) && (verified))
							GetDestination(decoded[1], decoded[2]);
						break;
					case "PUT":
						if((decoded.Length == 3) && (verified))
							GetDestination(decoded[1], decoded[2]);
						break;
					case "REC":
						if((decoded.Length == 2) && (verified))
							GetSource(decoded[1]);
						break;
					case "RSP":
						if (decoded.Length == 2)
							VerifyUser (decoded [1]);
						break;
					case "SHA":
						if ((decoded.Length == 2) && (verified))
							SaveShare (decoded [1]);
						break;
					case "VER":
						if(decoded.Length == 2)
							CheckVersion(decoded[1]);
						break;
					case "WRN":
						if((decoded.Length == 2) && (verified))
							GetDestination(filename, filesize);
						break;
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.Decode"); 
			}
		}

		/// <summary>
		/// Checks the client's version no.
		/// </summary>
		/// <param name="ver">the client's version</param>
		private void CheckVersion (string ver)
		{
			version = ver;
		}

		/// <summary>
		/// Creates a new thread to listen for messeages from this user
		/// </summary>
		public void StartListening ()
		{
			try
			{
				Thread listening = new Thread(new ThreadStart(Listen));		
				listening.IsBackground=true;
				listening.Start();
				run = true;
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.StartListening"); 
			}
		}

		/// <summary>
		/// Listens to the connection to the user and forwards incoming messages
		/// </summary>
		private void Listen()
		{
			try
			{
				using(reader)
				{
					while(run)
					{ 
						Decode(reader.ReadLine()); 
					}
				}
			}
			// if the client is no longer online, disconnect
			catch(IOException)
			{
				this.Disconnect();
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.Listen"); 
			}
		}

		/// <summary>
		/// Logs the user out
		/// </summary>
		public void Logout ()
		{
			Disconnect();
		}

		/// <summary>
		/// Calls a method to check if the user exists
		/// if the user exists, calls a method to challenge
		/// this user
		/// </summary>
		/// <param name="uName">username</param>
		private void UserExists (string uName)
		{
			User temp;
			userName = uName;

			temp = dbhandle.UserExists (userName);

			// if the user exists
			if (temp != null)
			{
				// if the user's account is not locked
				if (!temp.locked)
				{
					attempt = 0;
					this.password = temp.password;
					this.pubKey = temp.pubKey;
					this.share = temp.share;
					this.accessible = temp.accessible;
					this.available = temp.available;
					this.uid = temp.uid;
					Challenge();
				}
				else
				{
					try
					{
						writer.WriteLine("502");
						writer.Flush();
						if(LoggEvent!=null)
							LoggEvent(this.userName, false, "502");
					}
					catch(Exception ex)
					{
						MessageBox.Show(ex.Message,"Hub.User.UserExists"); 
					}
				}
			}
			else
			{
				attempt++;
				try
				{
					if (attempt < 3)
					{
						writer.WriteLine("401");
						writer.Flush();
						if(LoggEvent!=null)
							LoggEvent(this.userName, false, "401");
					}
					else
					{
						writer.WriteLine("501");
						writer.Flush();
						if(LoggEvent!=null)
							LoggEvent(this.userName, false, "501");
					}
				}
				catch(Exception ex)
				{
					MessageBox.Show(ex.Message,"Hub.User.UserExists"); 
				}
			}
		}

		/// <summary>
		/// Verifies the user by comparing the two hashed strings
		/// calls a method to add the users ip to the database 
		/// and sends an "OK" message to the user if they are equal
		/// </summary>
		/// <param name="response">the hashed string received from the user</param>
		private void VerifyUser (string response)
		{	
			if (verificationHash == response)
			{
				attempt = 0;
				verified = true;
				this.ip = (((IPEndPoint)connection.RemoteEndPoint).Address.ToString ());
				dbhandle.SaveIP(this.userName, this.ip);

				VerifyRequiredParameters();
			}
			else
			{
				try
				{
					attempt++;
					if (attempt < 3)
					{
						writer.WriteLine("402");
						writer.Flush();
						if(LoggEvent!=null)
							LoggEvent(this.userName, false, "402");
					}
					else
					{
						writer.WriteLine("502");
						writer.Flush();
						if(LoggEvent!=null)
							LoggEvent(this.userName, false, "502");
					}
				}
				catch(Exception ex)
				{
					MessageBox.Show(ex.Message,"Hub.User.VerifyUser"); 
				}
			}
		}

		/// <summary>
		/// Creates a timestamp and a random number and sends it to the client
		/// </summary>
		/// <param name="pass">the user's password</param>
		private void Challenge ()
		{
			string timeString;
			string trimmedTime;

			DateTime time = DateTime.Now;
			Random rand = new Random((int)time.Ticks);
			int no;

			no = rand.Next();

			timeString = time.ToString();
			trimmedTime = timeString.Remove(10, 1);

			string result;

			result = password + trimmedTime + no;
			verificationHash = HashString(result);
			
			try
			{
				writer.WriteLine("CHA " + trimmedTime + " " + no.ToString ());
				writer.Flush();
				if(LoggEvent!=null)
					LoggEvent(this.userName, false, "CHA " + trimmedTime + " " + no.ToString ());
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.Challenge"); 
			}

		}

		/// <summary>
		/// Calls a method to hash the user's password and 
		/// calls a method to add the user to the database
		/// </summary>
		public void AddNewUser()
		{
			string hashedPassword;
			hashedPassword = HashString(this.password);
			dbhandle.AddUser(this.userName, hashedPassword);
		}

		/// <summary>
		/// Calls a method to delete this user
		/// </summary>
		/// <param name="uN">username</param>
		public void DeleteUser(string uN)
		{
			dbhandle.DeleteUser(uN);
		}

		/// <summary>
		/// Hashes a string
		/// </summary>
		/// <param name="data">the string to hash</param>
		/// <returns>the bytearray converted to a string</returns>
		private string HashString(string data)
		{
			byte[] toHash = Encoding.ASCII.GetBytes(data);
			byte [] hashedResult;

			SHA1Managed hashAlg = new SHA1Managed();
			hashedResult = hashAlg.ComputeHash(toHash);
			return ByteArrayToString(hashedResult);
		}

		/// <summary>
		/// Takes a byte array and creates a string
		/// </summary>
		/// <param name="toConvert">the byte array to replace with a string</param>
		/// <returns>the bytearray converted to a string</returns>
		private string ByteArrayToString(byte [] toConvert)
		{
			StringBuilder sb = new StringBuilder(toConvert.Length);
			for (int i = 0; i < toConvert.Length - 1; i++)
			{
				
				sb.Append(toConvert[i].ToString("X"));
			}
			return sb.ToString();
		}

		/// <summary>
		/// Verifies that both share and key are ok when the user is logging in
		/// </summary>
		private void VerifyRequiredParameters()
		{
			try
			{
				if (CheckShare() == true)
				{
					if (CheckKey() == true)
					{
						writer.WriteLine("201");
						writer.Flush();
						if(LoggEvent != null)
							LoggEvent(this.userName, false, "201");
						if (AddUserEvent != null)
							AddUserEvent(this);
						SendAvailableSpace();
					}
					else
					{
						writer.WriteLine("302");
						writer.Flush();
						if(LoggEvent != null)
							LoggEvent(this.userName, false, "302");
					}
				}
				else
				{
					writer.WriteLine("301");
					writer.Flush();
					if(LoggEvent != null)
						LoggEvent(this.userName, false, "301");
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.VerifyRequiredParameters"); 
			}
		}

		/// <summary>
		/// Checks if the user has a share
		/// </summary>
		/// <returns>true or false</returns>
		private bool CheckShare()
		{
			if (share == 0)
			{
				return false;
			}
			else
			{
				return true;
			}
		}

		/// <summary>
		/// Checks if the user has sent its key
		/// </summary>
		/// <returns>true or false</returns>
		private bool CheckKey()
		{
			if (pubKey == "")
			{
				return false;
			}
			else
			{
				return true;
			}
		}

		/// <summary>
		/// Checks if the new share size is OK and if the user's public key exists
		/// </summary>
		/// <param name="size">the new size of the share</param>
		private void SaveShare(string size)
		{
			long tempSize = Int64.Parse(size);
			long tempAvail;
			long tempAccess;

			try
			{
				// if new size is larger than used space and larger than stored
				if (tempSize > (share - available) && tempSize > (share - accessible))
				{
					tempAvail = tempSize - share + available;
					tempAccess = tempSize - share + accessible;
					dbhandle.SaveShare(this.userName, tempSize, tempAvail, tempAccess);
					// if it's a new share
					if(share == 0)
					{
						this.share = tempSize;
						this.available = tempAvail;
						this.accessible = tempAccess;
						// if the user has a key
						if (CheckKey() == true)
						{
							writer.WriteLine("208");
							writer.Flush();
							if(LoggEvent != null)
								LoggEvent(this.userName, false, "208");
							if (AddUserEvent != null)
								AddUserEvent(this);
							SendAvailableSpace();
						}
						// if the user hasn't got a key
						else
						{
							writer.WriteLine("302");
							writer.Flush();
							if(LoggEvent != null)
								LoggEvent(this.userName, false, "302");
						}
					}
					// if it's a changed share
					else
					{
						this.share = tempSize;
						this.available = tempAvail;
						this.accessible = tempAccess;
						if (CheckKey() == true)
						{
							writer.WriteLine("208");
							writer.Flush();
							if(LoggEvent != null)
								LoggEvent(this.userName, false, "208");
						}
					}			
				}
				// if the share is too small
				else
				{
					writer.WriteLine("404");
					writer.Flush();
					if(LoggEvent != null)
						LoggEvent(this.userName, false, "404");
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.SaveShare"); 
			}
		}

		/// <summary>
		/// Calls a method to save the user's public key
		/// Sends a message to the user that the key is received
		/// </summary>
		/// <param name="key">the public key</param>
		private void SaveKey(string key)
		{
			dbhandle.SaveKey(userName, key);

			try
			{
				if(pubKey == "")
				{
					this.pubKey = key;
					writer.WriteLine("209");
					writer.Flush();
					if(LoggEvent != null)
						LoggEvent(this.userName, false, "209");
					if (AddUserEvent != null)
						AddUserEvent(this);
					SendAvailableSpace();
				}
				else
				{
					this.pubKey = key;
					writer.WriteLine("209");
					writer.Flush();
					if(LoggEvent != null)
						LoggEvent(this.userName, false, "209");
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.SaveKey"); 
			}
		}

		/// <summary>
		/// Closes the socket and streams for this user
		/// if the connected user has disconnected
		/// Sets the bool attributes to false
		/// </summary>
		public void Disconnect()
		{
			try
			{
				verified = false;
				run = false;
				Thread.Sleep(1024);
				if(this.connection != null && this.connection.Connected)
				{
					this.connection.Shutdown(SocketShutdown.Both);
					this.connection.Close();
				}
				if(writer != null)
					writer.Close();
				if(reader != null)
					reader.Close();
				if (RemoveUserEvent != null)
					RemoveUserEvent(this);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.Disconnect"); 
			}
		}

		/// <summary>
		/// Checks if there's enough space for the requested backup
		/// this method should also consider the estimated size
		/// of the index file
		/// </summary>
		/// <param name="s">size of the backup</param>
		private void Backup(string s)
		{
			double size = Int64.Parse(s);

			try
			{
				if(size < accessible)
				{
					writer.WriteLine("202");
					writer.Flush();
					if(LoggEvent!=null)
						LoggEvent(this.userName, false, "202");
				}
				else
				{
					writer.WriteLine("403");
					writer.Flush();
					if(LoggEvent!=null)
						LoggEvent(this.userName, false, "403");
				}
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.Backup"); 
			}
		}

		/// <summary>
		/// Overloaded; this one takes two strings
		/// Calls a method to get a destination for the backup
		/// Sends information to the destination user about the backup
		/// </summary>
		/// <param name="n">filename</param>
		/// <param name="s">filesize</param>
		private void GetDestination(string n, string s)
		{
			User destination = null;
			action = 1;

			filename = n;
			filesize = Int64.Parse(s);
			// temporarily change accessible space
			accessible = accessible - filesize;

			if (CalculateDestinationEvent != null)
				destination = CalculateDestinationEvent(this);

			if (destination != null)
			{
				try
				{
					destination.writer.WriteLine("RCV " + this.ip + " " + this.userName + " " + this.pubKey);
					destination.writer.Flush();
					if(LoggEvent != null)
						LoggEvent(destination.userName, false, "RCV " + this.ip + " " + this.userName + " " + this.pubKey);
					dbhandle.AddFile(uid, filename, filesize, destination.uid);
					// temporarily change destination's available space
					destination.available = destination.available - filesize;					
				}
				catch(IOException)
				{
					// change back destination's available space
					destination.available = destination.available + filesize;
					// delete the post in the database
					dbhandle.DeleteFile(uid, filename, destination.uid);
					GetDestination(filename, filesize);
				}
				catch(Exception ex)
				{
					MessageBox.Show(ex.Message,"Hub.User.GetDestination"); 
				}
			}
			else
			{
				try
				{
					writer.WriteLine("406");
					writer.Flush();
					if(LoggEvent != null)
						LoggEvent(this.userName, false, "406");
				}
				catch (IOException ex)
				{
					MessageBox.Show(ex.Message,"Hub.User.GetDestination");
				}
			}
		}

		/// <summary>
		/// Overloaded; this one takes a double
		/// Calls a method to get a destination for the backup
		/// Sends information to the destination user about the backup
		/// </summary>
		/// <param name="n">filename</param>
		/// <param name="s">filesize</param>
		private void GetDestination(string n, long f)
		{
			User destination = null;

			if (CalculateDestinationEvent != null)
				destination = CalculateDestinationEvent(this);

			if (destination != null)
			{
				try
				{
					destination.writer.WriteLine("RCV " + this.ip + " " + this.userName + " " + this.pubKey);
					destination.writer.Flush();
					if(LoggEvent != null)
						LoggEvent(destination.userName, false, "RCV " + this.ip + " " + this.userName + " " + this.pubKey);
					dbhandle.AddFile(uid, filename, filesize, destination.uid);
					// temporarily change destination's available space
					destination.available = destination.available - filesize;		
				}
				catch(IOException)
				{
					// change back destination's available space
					destination.available = destination.available + filesize;
					// delete the post in the database
					dbhandle.DeleteFile(uid, filename, destination.uid);
					GetDestination(filename, filesize);
				}
				catch(Exception ex)
				{
					MessageBox.Show(ex.Message,"Hub.User.GetDestination"); 
				}
			}
			else
			{
				try
				{
					writer.WriteLine("406");
					writer.Flush();
					if(LoggEvent != null)
						LoggEvent(this.userName, false, "406");
				}
				catch (IOException ex)
				{
					MessageBox.Show(ex.Message,"Hub.User.GetDestination");
				}
			}

		}

		/// <summary>
		/// Sends the available space to this user
		/// </summary>
		private void SendAvailableSpace()
		{
			action = 0;

			try
			{
				writer.WriteLine("ACS " + this.accessible);
				writer.Flush();
				if(LoggEvent != null)
					LoggEvent(userName, false, "ACS " + this.accessible);
			}
			catch(Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.SendAvailable"); 
			}
		}

		/// <summary>
		/// Calls the database to:
		/// set the file as transferred
		/// change the source's accesible space
		/// change the destination's available space
		/// </summary>
		/// <param name="fName">the file that's transferred</param>
		/// <param name="destName">the destination's username</param>
		private void FileDone (string fName, string destName)
		{
			User destination = null;

			if (GetClientEvent != null)
				destination = GetClientEvent(destName);

			dbhandle.FileDone(this.uid, fName, destName);
			dbhandle.ChangeAccessible(this.userName, this.accessible);
			dbhandle.ChangeAvailable(destName, destination.available);

			filename = "";

			try
			{
				writer.WriteLine("215");
				writer.Flush();
				if(LoggEvent != null)
					LoggEvent(userName, false, "215");
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.FileDone"); 
			}
		}

		/// <summary>
		/// Gets the user storing the file at recovery time
		/// </summary>
		/// <param name="userName">username for storing user</param>
		private void GetSource(string userName)
		{
			User source = null;
			action = 2;

			if (GetClientEvent != null)
				source = GetClientEvent(userName);

			try
			{
				if (source != null)
				{
					source.writer.WriteLine("LET" + " " + this.ip + " " + this.userName + " " + this.pubKey);
					source.writer.Flush();
					if(LoggEvent != null)
						LoggEvent(source.userName, false, "LET" + " " + this.ip + " " + this.userName + " " + this.pubKey);
				}
				else
				{
					writer.WriteLine("405");
					writer.Flush();
					if(LoggEvent != null)
						LoggEvent(userName, false, "405");
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.GetSource"); 
			}
		}

		/// <summary>
		/// Gets the client who is requesting backup or recovery
		/// </summary>
		/// <param name="userName">the requesting client</param>
		private void GetClient(string userName)
		{
			User reqClient = null;

			if (GetClientEvent != null)
				reqClient = GetClientEvent(userName);

			if (reqClient != null)
			{
				// if the requesting user is backing up, call requesting user's SendDestination
				if (reqClient.action == 1)
					reqClient.SendDestination(this);
				// if the requesting user is recovering, call requesting user's SendSource
				if (reqClient.action == 2)
					reqClient.SendSource(this);
			}
			// if the requesting client is not online
			else
			{
				try
				{
					writer.WriteLine("STP " + reqClient.userName);
					writer.Flush();
				}
				catch (Exception ex)
				{
					MessageBox.Show(ex.Message,"Hub.User.GetClient"); 
				}
				if(LoggEvent != null)
					LoggEvent(userName, false, "STP " + reqClient.userName);
			}
		}

		/// <summary>
		/// Sends information about the storing client to the requesting client
		/// at recovery time
		/// </summary>
		/// <param name="source">the user storing the file</param>
		private void SendSource(User source)
		{
			try
			{
				writer.WriteLine("SRC " + source.ip + " " + source.userName + " " + source.pubKey);
				writer.Flush();
				if(LoggEvent != null)
						LoggEvent(userName, false, "SRC " + source.ip + " " + source.userName + " " + source.pubKey);
			}				
			catch (IOException ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.SendSource");
			}
		}

		/// <summary>
		/// Sends information about the destination client to the requesting client
		/// at backup time
		/// </summary>
		/// <param name="dest">the receiving user</param>
		private void SendDestination(User dest)
		{
			try
			{
				writer.WriteLine("DST " + filename + " " + dest.ip + " " + dest.userName + " " + dest.pubKey);
				writer.Flush();
				if(LoggEvent != null)
					LoggEvent(userName, false, "DST " + filename + " " + dest.ip + " " + dest.userName + " " + dest.pubKey);
			}
			catch (IOException ex)
			{
				MessageBox.Show(ex.Message,"Hub.User.SendDestination");
			}
		}
	}
}
